#include "stdwtl.h"
#include "objectstree.h"
#include "./perf/counterhelpdb.h"
#include "./perf/perfobjects.h"

namespace
{
	// ,   
	const int image_x     = 16;
	const int image_y     = 16;
	const int image_count = 5;
};

//////////////////////////////////////////////////////////////////////
//
//  
//
//////////////////////////////////////////////////////////////////////

CObjectsTree::CObjectsTree() : _sort_insertion(true)
{
}

CObjectsTree::~CObjectsTree()
{
	_image.Destroy();
}

bool CObjectsTree::Create(HWND parent)
{
	CreateImageList();

	HWND hTree;

	hTree = CTreeViewCtrl::Create( 
						parent,
						NULL, 
						NULL, 
						WS_CHILD | WS_VISIBLE | WS_BORDER
					  | TVS_HASBUTTONS | TVS_DISABLEDRAGDROP | TVS_HASLINES  
					  | TVS_LINESATROOT | TVS_SHOWSELALWAYS,
						WS_EX_CLIENTEDGE);

	if (NULL == hTree)
	{
		return false;
	}

	SetImageList(_image, TVSIL_NORMAL);

	return true;
}

void CObjectsTree::Clear()
{
	DeleteAllItems();
}

//
//   
//
void CObjectsTree::Fill(int full_size, const void* d)
{
	Clear();

	if (full_size == 0) return;
	if (d == 0) return;

	perf_data_block data(d);

	if (data._data->TotalByteLength > full_size) return;

	HTREEITEM hSystem;
	
	hSystem = InsertW(
				TVI_ROOT, 
				data.name(), 
				data.name_len(),
				system_id, 
				data._data);

	perf_data_block::iterator i = data.get_iterator();
	
	while (i.is_any())
	{
		const PERF_OBJECT_TYPE* object = i.get_next();

		FillObject(hSystem, object);
	}
}

//
//     
void CObjectsTree::GetCurrentItem(int* type, DWORD* data, DWORD* parent_type, DWORD* parent_data)
{
	HTREEITEM hCurr = GetSelectedItem();

	if (hCurr == NULL) 
	{
		//  
		*type = none;
		return;
	}

	TVITEM item = {0};

	item.mask  = TVIF_HANDLE | TVIF_PARAM | TVIF_IMAGE;
	item.hItem = hCurr;

	GetItem(&item);

	*type = item.iImage; //        
	*data = item.lParam;

	if (*type != counter_id)
	{
		//     
		//      
		//  .    .
		*parent_data = 0;
		return;
	}

	HTREEITEM hParent = GetParentItem(hCurr);

	if (hParent == NULL)
	{
		*parent_type = none;
		*parent_data = 0;
		return;
	}

	item.hItem = hParent;

	GetItem(&item);
		
	*parent_type = item.iImage;
	*parent_data = item.lParam;
}

///////////////////////////////////////////////////////////////
//
//  
//
///////////////////////////////////////////////////////////////

bool CObjectsTree::CreateImageList()
{
	_image.Create(image_x, image_y, 0, image_count, 0);

	HBITMAP hBmp;

	hBmp = LoadBitmap(
		_Module.GetResourceInstance(), 
		MAKEINTRESOURCE(IDB_SYSTEM));

	_image.Add(hBmp);

	DeleteObject(hBmp);

	hBmp = LoadBitmap(
		_Module.GetResourceInstance(), 
		MAKEINTRESOURCE(IDB_OBJECT));
	
	_image.Add(hBmp);

	DeleteObject(hBmp);

	hBmp = LoadBitmap(
		_Module.GetResourceInstance(), 
		MAKEINTRESOURCE(IDB_MULTI_OBJECT));
	
	_image.Add(hBmp);

	DeleteObject(hBmp);

	hBmp = LoadBitmap(
		_Module.GetResourceInstance(), 
		MAKEINTRESOURCE(IDB_INSTANCE));
	
	_image.Add(hBmp);

	DeleteObject(hBmp);

	hBmp = LoadBitmap(
		_Module.GetResourceInstance(), 
		MAKEINTRESOURCE(IDB_FAKE_INSTANCE));
	
	_image.Add(hBmp);

	DeleteObject(hBmp);

	hBmp = LoadBitmap(
		_Module.GetResourceInstance(), 
		MAKEINTRESOURCE(IDB_COUNTER));
	
	_image.Add(hBmp);

	DeleteObject(hBmp);

	return true;
}

//
//     
void CObjectsTree::FillObject(HTREEITEM parent, const PERF_OBJECT_TYPE* o)
{
	perf_object_type object(o);
	HTREEITEM hObject;

	if (!object.is_instances())
	{
		//    
		hObject = InsertTitle(parent, object._data->ObjectNameTitleIndex, object_id, object._data);
		FillCounters(hObject, object._data);
		return;
	}

	// 
	hObject = InsertTitle(
					parent, 
					object._data->ObjectNameTitleIndex, 
					multi_object_id, 
					object._data);

	perf_object_type::instance_iterator i = object.get_instance_iterator();
	
	HTREEITEM hInstance;

	if (!i.is_any())
	{
		//    
		hInstance = Insert(hObject, TEXT("FAKE"), fake_instance_id, NULL);
		FillCounters(hInstance, object._data);
		return;
	}

	//   
	while (i.is_any())
	{
		perf_instance_definition instance = i.get_next();

		if (instance.name())
		{
			//  
			hInstance = InsertW(
							hObject, 
							instance.name(),
							instance.name_len(),
							instance_id, 
							instance._data);
		}
		else
		{
			//  ID
			hInstance = InsertID(
							hObject, 
							instance._data->UniqueID,
							instance_id, 
							instance._data);
		}

		FillCounters(hInstance, object._data);
	}
}

//
//     
void CObjectsTree::FillCounters(HTREEITEM parent, const PERF_OBJECT_TYPE* o)
{
	perf_object_type object(o);
	perf_object_type::counter_iterator i = object.get_counter_iterator();

	while (i.is_any())
	{
		const PERF_COUNTER_DEFINITION* counter = i.get_next();
		InsertTitle(parent, counter->CounterNameTitleIndex, counter_id, counter);
	}
}

//
//     
HTREEITEM CObjectsTree::InsertW(HTREEITEM parent, const WCHAR* nameW, int len, int img, const void* data)
{
#ifndef UNICODE
	
	int req = WideCharToMultiByte(CP_ACP, 0, nameW, len, NULL, 0, 0, 0); 

	char* name = (char*) malloc(req + 1);
	name[0] = 0;

	WideCharToMultiByte(CP_ACP, 0, nameW, len, name, req, 0, 0); 
	name[req] = 0;

#else  // UNICODE

	WCHAR* name = (WCHAR*) malloc((len + 1) * sizeof(WCHAR));
	memcpy(name, nameW, len * sizeof(WCHAR));
	name[len] = 0;

#endif // UNICODE

	HTREEITEM res = Insert(parent, name, img, data);

	free(name);

	return res;
}

//
//     
HTREEITEM CObjectsTree::InsertTitle(HTREEITEM parent, int name, int img, const void* data)
{
	return Insert(parent, CounterHelpDB::GetName(name), img, data);
}

//
//     
HTREEITEM CObjectsTree::InsertID(HTREEITEM parent, int id, int img, const void* data)
{
	CString text;
	text.Format(TEXT("%d"), id);

	return Insert(parent, text, img, data);
}

//
//     
HTREEITEM CObjectsTree::Insert(HTREEITEM parent, const TCHAR* name, int img, const void* data)
{
	TV_INSERTSTRUCT item = {0};

	item.hParent = parent;
	
	if ((parent == NULL) || (parent == TVI_ROOT))
	{
		item.hInsertAfter = TVI_ROOT;
	}
	else
	{
		if (_sort_insertion)
		{
			//   
			item.hInsertAfter = TVI_SORT;
		}
		else
		{
			//    
			item.hInsertAfter = TVI_LAST;
		}
	}

	item.item.mask           = TVIF_TEXT | TVIF_IMAGE | TVIF_SELECTEDIMAGE | TVIF_PARAM;
	item.item.pszText        = (TCHAR*)name;
	item.item.iImage         = img;
	item.item.iSelectedImage = img;
	item.item.lParam         = (LPARAM)data;

	return InsertItem(&item);
}


/////////////////////////////////////////////////////////